#!/usr/bin/env python
# -*- coding: utf-8 -*-
###############################################################################
# Filename:          AtmosphericExtinction.py
# Copyright          2012, Clément Buton, Yannick Copin
# Author:            Clément Buton <buton@physik.uni-bonn.de>
# Author:            $Author: ycopin $
# Version:           $Revision: 1.9 $
# Modified at:       $Date: 2017/01/04 18:16:33 $
# $Id: atmosphericExtinction.py,v 1.9 2017/01/04 18:16:33 ycopin Exp $
###############################################################################

import os
import numpy as N
import pyExtinction.AtmosphericExtinction as AtmExt

__author__ = "Yannick Copin <y.copin@ipnl.in2p3.fr>, " \
             "Clément Buton <c.buton@ipnl.in2p3.fr>"
__version__ = '$Id: atmosphericExtinction.py,v 1.9 2017/01/04 18:16:33 ycopin Exp $'

if __name__ == "__main__":

    import optparse

    # Options =================================================================

    usage = "usage: %prog [options]"

    description = """\
Compute and/or plot atmospheric extinction from physical parameters,
as described in Buton et al. (2013A&A...549A...8B), *Atmospheric
extinction properties above Mauna Kea from the Nearby Supernova
Factory spectro-photometric data set*.

Default extinction parameters correspond to mean Mauna-Kea summit
conditions.  Each extinction parameter can be specified using option
'--param value stderr'.
"""
    epilog = """\
If you have found this software useful for your research, we would
appreciate a reference to *the atmospheric extinction of Buton et
al. (2013)*."""

    parser = optparse.OptionParser(usage, version=__version__,
                                   description=description,
                                   epilog=epilog)

    # Rayleigh parameters ------------------------------

    rayleigh = optparse.OptionGroup(parser, "Rayleigh parameters")

    rayleigh.add_option('-p', '--pressure', type=float, nargs=2,
                        help="Surface pressure [%default mbar]",
                        default=(616., 2.))

    parser.add_option_group(rayleigh)

    # Ozone parameters ------------------------------

    ozone = optparse.OptionGroup(parser, "Ozone parameters")

    ozone.add_option('-t', '--o3template', metavar='FILE.fits',
                     help="Ozone transmission template",
                     default=AtmExt.O3Template)
    ozone.add_option('-c', '--o3column', type=float, nargs=2,
                     help="Ozone column density [%default DU]",
                     default=(257., 23.))

    parser.add_option_group(ozone)

    # Aerosol parameters ------------------------------

    aerosols = optparse.OptionGroup(parser, "Aerosol parameters")

    aerosols.add_option('-d', '--depth', type=float, nargs=2,
                        help="Aerosol optical depth at 1 micron [%default]",
                        default=(7.6e-3, 1.4e-3))
    aerosols.add_option('-a', '--angstrom', type=float, nargs=2,
                        help="Aerosol Angstrom exponent [%default]",
                        default=(1.26, 1.33))

    parser.add_option_group(aerosols)

    # Output ------------------------------

    outputs = optparse.OptionGroup(parser, "Output parameters")

    outputs.add_option('-w', '--wrange', metavar='START,END,STEP',
                       type=float, nargs=3,
                       help="Wavelength range [%default A]",
                       default=(3200, 10001, 10))

    outputs.add_option('-o', '--output', metavar="FILE.txt|fits",
                       help="Output table name (.txt|.fits)",
                       default="atmosphericExtinction.txt")

    outputs.add_option("-P", "--plot", action='store_true',
                       help="Interactive plot")
    outputs.add_option("-T", "--transmission", action='store_true',
                       help="Plot transmission rather than extinction")

    parser.add_option_group(outputs)

    # Check for options consistency ============================================

    opts, args = parser.parse_args()

    # Wavelength range
    wStart, wEnd, wStep = opts.wrange
    if (wStart > wEnd) or (wStep > (wEnd - wStart)):
        parser.error("Invalid wavelength range '--wrange %s'" %
                     (','.join(opts.wrange)))

    # Output
    if opts.output:                     # Decipher output format
        basename, outformat = os.path.splitext(opts.output)
        if outformat.lower() not in ('.fits', '.txt'):
            parser.error("Unknown output format '%s'" % outformat)
        outformat = outformat[1:].lower()  # Remove leading dot

    # ==========================================================================

    # Inputs parameters ------------------------------

    # Wavelength domain
    lbda = N.arange(wStart, wEnd, wStep, dtype=float)

    # Extinction parameters
    o3, do3 = opts.o3column     # Ozone column density [DU]
    ang, dang = opts.angstrom   # Ångström exponent
    tau, dtau = opts.depth      # Aerosol optical depth at ref. wavelength
    p, dp = opts.pressure       # Surface pressure [mbar]

    # Extinction model ------------------------------

    extModel = AtmExt.ExtinctionModel(lbda, opts.o3template)
    ext = extModel.extinction([ p, o3, tau, ang],
                              [dp, do3, dtau, dang], components=True)
    desc = str(extModel)
    print desc

    # Create output tables ------------------------------

    if opts.output:
        print "Saving extinction table '%s'..." % opts.output
        extModel.write(opts.output, ext, outformat)

    # Create output graphics ------------------------------

    if opts.plot:
        print "Generating interactive figure..."
        import matplotlib.pyplot as P

        ax = extModel.plot(ext, components=True, transmission=opts.transmission)
        P.show()

    print epilog
